@@ -71,7 +71,7 @@ namespace {
   bool isRcVisardDevice(const std::string& vendor, const std::string& model)
   {
     bool isKuka3DPerception = vendor.find("KUKA") != std::string::npos && model.find("3d_perception") != std::string::npos;
-    bool isRcVisard = vendor.find("Roboception") != std::string::npos && model.find("rc_visard") != std::string::npos;
+    bool isRcVisard = model.find("rc_visard") != std::string::npos;
     return isKuka3DPerception || isRcVisard;
   }
 
@@ -98,7 +98,21 @@ namespace {
     }
     return device_ids;
   }
-}
+
+  void split(std::vector<std::string>& list, const std::string& s, char delim, bool skip_empty = true)
+  {
+    std::stringstream in(s);
+    std::string elem;
+
+    while (getline(in, elem, delim))
+    {
+      if (!skip_empty || elem.size() > 0)
+      {
+        list.push_back(elem);
+      }
+    }
+  }
+}  // namespace
 
 namespace rc
 {
@@ -131,12 +145,12 @@ DeviceNodelet::DeviceNodelet()
   reconfig = 0;
   dev_supports_gain = false;
   dev_supports_color = false;
-  dev_supports_chunk_data = false;
   dev_supports_wb = false;
   dev_supports_depth_acquisition_trigger = false;
   perform_depth_acquisition_trigger = false;
   iocontrol_avail = false;
   dev_supports_double_shot = false;
+  dev_supports_depth_exposure_adapt_timeout = false;
   level = 0;
 
   stopImageThread = imageRequested = imageSuccess = false;
@@ -345,26 +359,30 @@ void DeviceNodelet::keepAliveAndRecoverFromFails()
         rcgdev->open(access_id);
         rcgnodemap = rcgdev->getRemoteNodeMap();
 
-        // in newer rc_visard image versions, we can check via genicam if rc_visard is ready
+        // ensure that device version >= 20.04
 
-        try {
-          bool isReady = rcg::getBoolean(rcgnodemap, "RcSystemReady", true, true);
-          if (!isReady)
-          {
-            ROS_INFO_STREAM("rc_visard_driver: rc_visard device not yet ready (GEV). Trying again...");
-            continue;  // to next trial!
-          }
-        } catch (std::invalid_argument& e)
+        dev_version = rcg::getString(rcgnodemap, "DeviceVersion");
+
+        std::vector<std::string> list;
+        split(list, dev_version, '.');
+
+        if (list.size() < 3 || std::stoi(list[0]) < 20 || (std::stoi(list[0]) == 20 && std::stoi(list[1]) < 4))
         {
-          // genicam feature is not implemented in this version of rc_visard's firmware,
-          // so we have to assume rc_visard is ready and continue
+          throw std::invalid_argument("Device version must be 20.04 or higher: " + dev_version);
+        }
+
+        // check if device is ready
+
+        if (!rcg::getBoolean(rcgnodemap, "RcSystemReady", true, true))
+        {
+          ROS_INFO_STREAM("rc_visard_driver: rc_visard device not yet ready (GEV). Trying again...");
+          continue;  // to next trial!
         }
 
         // extract some diagnostics data from device
         dev_serialno = rcg::getString(rcgnodemap, "DeviceID", true);
         dev_macaddr = rcg::getString(rcgnodemap, "GevMACAddress", true);
         dev_ipaddr = rcg::getString(rcgnodemap, "GevCurrentIPAddress", true);
-        dev_version = rcg::getString(rcgnodemap, "DeviceVersion", true);
         gev_userid = rcg::getString(rcgnodemap, "DeviceUserID", true);
         gev_packet_size = rcg::getString(rcgnodemap, "GevSCPSPacketSize", true);
 
@@ -564,8 +582,16 @@ void DeviceNodelet::initConfiguration(const std::shared_ptr<GenApi::CNodeMapRef>
     if (format == "YCbCr411_8")
     {
       dev_supports_color = true;
+      color_format = "YCbCr411_8";
       break;
     }
+    if (format == "RGB8")
+    {
+      dev_supports_color = true;
+      color_format = "RGB8";
+      break;
+    }
+
   }
 
   // get optional white balancing values (only for color camera)
@@ -617,15 +643,7 @@ void DeviceNodelet::initConfiguration(const std::shared_ptr<GenApi::CNodeMapRef>
   cfg.depth_maxdepth = rcg::getFloat(nodemap, "DepthMaxDepth", 0, 0, true);
   cfg.depth_maxdeptherr = rcg::getFloat(nodemap, "DepthMaxDepthErr", 0, 0, true);
 
-  cfg.ptp_enabled = rcg::getBoolean(nodemap, "GevIEEE1588", false);
-
-  // fix for rc_visard < 1.5
-
-  if (cfg.depth_quality[0] == 'S')
-  {
-    cfg.depth_quality = "High";
-    cfg.depth_static_scene = true;
-  }
+  cfg.ptp_enabled = rcg::getBoolean(nodemap, "PtpEnable", false);
 
   // check for stereo_plus license
 
@@ -660,10 +678,6 @@ void DeviceNodelet::initConfiguration(const std::shared_ptr<GenApi::CNodeMapRef>
 
   try
   {
-    // disable filtering images on the sensor
-
-    rcg::setEnum(nodemap, "AcquisitionAlternateFilter", "Off", false);
-
     // get current values
 
     rcg::setEnum(nodemap, "LineSelector", "Out1", true);
@@ -694,11 +708,19 @@ void DeviceNodelet::initConfiguration(const std::shared_ptr<GenApi::CNodeMapRef>
   // enabling chunk data for getting live camera/image parameters
   // (e.g. line status, current gain and exposure, etc.)
 
-  dev_supports_chunk_data = rcg::setBoolean(nodemap, "ChunkModeActive", true, false);
-  if (!dev_supports_chunk_data)
-  {
-    ROS_WARN("rc_visard_driver: rc_visard has an older firmware that does not support chunk data.");
+  rcg::setEnum(nodemap, "AcquisitionAlternateFilter", "Off", false);
+  rcg::setEnum(nodemap, "AcquisitionMultiPartMode", "SingleComponent", true);
+  rcg::setBoolean(nodemap, "ChunkModeActive", true, true);
 
+  try
+  {
+    cfg.depth_exposure_adapt_timeout = rcg::getFloat(nodemap, "DepthExposureAdaptTimeout", 0, 0, true);
+    dev_supports_depth_exposure_adapt_timeout = true;
+  }
+  catch (const std::exception&)
+  {
+    dev_supports_depth_exposure_adapt_timeout = false;
+    NODELET_WARN("rc_visard_driver: rc_visard has an older firmware, depth_exposure_adapt_timeout is not available.");
   }
 
   // try to get ROS parameters: if parameter is not set in parameter server, default to current sensor configuration
@@ -727,6 +749,7 @@ void DeviceNodelet::initConfiguration(const std::shared_ptr<GenApi::CNodeMapRef>
   pnh.param("depth_mindepth", cfg.depth_mindepth, cfg.depth_mindepth);
   pnh.param("depth_maxdepth", cfg.depth_maxdepth, cfg.depth_maxdepth);
   pnh.param("depth_maxdeptherr", cfg.depth_maxdeptherr, cfg.depth_maxdeptherr);
+  pnh.param("depth_exposure_adapt_timeout", cfg.depth_exposure_adapt_timeout, cfg.depth_exposure_adapt_timeout);
   pnh.param("ptp_enabled", cfg.ptp_enabled, cfg.ptp_enabled);
   pnh.param("out1_mode", cfg.out1_mode, cfg.out1_mode);
   pnh.param("out2_mode", cfg.out2_mode, cfg.out2_mode);
@@ -757,6 +780,7 @@ void DeviceNodelet::initConfiguration(const std::shared_ptr<GenApi::CNodeMapRef>
   pnh.setParam("depth_mindepth", cfg.depth_mindepth);
   pnh.setParam("depth_maxdepth", cfg.depth_maxdepth);
   pnh.setParam("depth_maxdeptherr", cfg.depth_maxdeptherr);
+  pnh.setParam("depth_exposure_adapt_timeout", cfg.depth_exposure_adapt_timeout);
   pnh.setParam("ptp_enabled", cfg.ptp_enabled);
   pnh.setParam("out1_mode", cfg.out1_mode);
   pnh.setParam("out2_mode", cfg.out2_mode);
@@ -834,7 +858,7 @@ void DeviceNodelet::reconfigure(rc_visard_driver::rc_visard_driverConfig& c, uin
     if (c.out1_mode != "Low" && c.out1_mode != "High" && c.out1_mode != "ExposureActive" &&
         c.out1_mode != "ExposureAlternateActive")
     {
-      c.out1_mode = "ExposureActive";
+      c.out1_mode = "Low";
     }
 
     if (c.out2_mode != "Low" && c.out2_mode != "High" && c.out2_mode != "ExposureActive" &&
@@ -845,10 +869,42 @@ void DeviceNodelet::reconfigure(rc_visard_driver::rc_visard_driverConfig& c, uin
   }
   else
   {
-    c.out1_mode = "ExposureActive";
+    c.out1_mode = "Low";
     c.out2_mode = "Low";
   }
 
+  if (!dev_supports_depth_exposure_adapt_timeout)
+  {
+    c.depth_exposure_adapt_timeout = 0.0;
+    l &= ~268435456;
+  }
+
+  // if out1_mode or out2_mode are changed, immediately set new value here
+  // If we do this in the grabbing thread we have a race condition and another parameter update
+  if (l & 262144)
+  {
+    l &= ~262144;
+
+    if (iocontrol_avail && rcgnodemap)
+    {
+      rcg::setEnum(rcgnodemap, "LineSelector", "Out1", false);
+      rcg::setEnum(rcgnodemap, "LineSource", c.out1_mode.c_str(), false);
+      ROS_DEBUG_STREAM("Set LineSource Out1 to " << c.out1_mode);
+    }
+  }
+
+  if (l & 524288)
+  {
+    l &= ~524288;
+
+    if (iocontrol_avail && rcgnodemap)
+    {
+      rcg::setEnum(rcgnodemap, "LineSelector", "Out2", false);
+      rcg::setEnum(rcgnodemap, "LineSource", c.out2_mode.c_str(), false);
+      ROS_DEBUG_STREAM("Set LineSource Out2 to " << c.out2_mode);
+    }
+  }
+
   // If camera_exp_auto (ExposureAuto) is changed, immediately set new value here
   // so that ExposureTime and Gain can be read and published again.
   // If we do this in the grabbing thread we have a race condition and another parameter update
@@ -1162,36 +1218,23 @@ void setConfiguration(const std::shared_ptr<GenApi::CNodeMapRef>& nodemap,
         ROS_DEBUG_STREAM("Set DepthMaxDepthErr to " << cfg.depth_maxdeptherr);
       }
 
+      if (lvl & 268435456)
+      {
+        lvl &= ~268435456;
+        rcg::setFloat(nodemap, "DepthExposureAdaptTimeout", cfg.depth_exposure_adapt_timeout, false);
+        ROS_DEBUG_STREAM("Set DepthExposureAdaptTimeout to " << cfg.depth_exposure_adapt_timeout);
+      }
+
       if (lvl & 131072)
       {
         lvl &= ~131072;
-        rcg::setBoolean(nodemap, "GevIEEE1588", cfg.ptp_enabled, true);
-        ROS_DEBUG_STREAM("Set GevIEEE1588 to " << cfg.ptp_enabled);
+        rcg::setBoolean(nodemap, "PtpEnable", cfg.ptp_enabled, true);
+        ROS_DEBUG_STREAM("Set PtpEnable to " << cfg.ptp_enabled);
       }
 
-      if (lvl & 262144)
-      {
-        lvl &= ~262144;
+      // lvl 262144 (out1_mode) is immediately set in dynamic reconfigure callback
 
-        if (iocontrol_avail)
-        {
-          rcg::setEnum(nodemap, "LineSelector", "Out1", true);
-          rcg::setEnum(nodemap, "LineSource", cfg.out1_mode.c_str(), true);
-          ROS_DEBUG_STREAM("Set LineSource Out1 to " << cfg.out1_mode);
-        }
-      }
-
-      if (lvl & 524288)
-      {
-        lvl &= ~524288;
-
-        if (iocontrol_avail)
-        {
-          rcg::setEnum(nodemap, "LineSelector", "Out2", true);
-          rcg::setEnum(nodemap, "LineSource", cfg.out2_mode.c_str(), true);
-          ROS_DEBUG_STREAM("Set LineSource Out2 to " << cfg.out2_mode);
-        }
-      }
+      // lvl 524288 (out2_mode) is immediately set in dynamic reconfigure callback
     }
     catch (const std::exception& ex)
     {
@@ -1253,21 +1296,66 @@ int enable(const std::shared_ptr<GenApi::CNodeMapRef>& nodemap, const char* comp
   return 0;
 }
 
+namespace
+{
+template <class T>
+inline rc_common_msgs::KeyValue getKeyValue(const char* key, T value)
+{
+  rc_common_msgs::KeyValue ret;
+
+  ret.key = key;
+  ret.value = std::to_string(value);
+
+  return ret;
+}
+
+inline rc_common_msgs::KeyValue getKeyValueString(const char* key, const std::string& value)
+{
+  rc_common_msgs::KeyValue ret;
+
+  ret.key = key;
+  ret.value = value;
+
+  return ret;
+}
+
+}  // namespace
+
 rc_common_msgs::CameraParam extractChunkData(const std::shared_ptr<GenApi::CNodeMapRef>& rcgnodemap)
 {
   rc_common_msgs::CameraParam cam_param;
 
-  // get list of all available IO lines and iterate over them to get their LineSource values
+  // get list of all available IO lines and iterate over them
+  // to get input/output masks and LineSource values for outputs
+
   std::vector<std::string> lines;
-  rcg::getEnum(rcgnodemap, "ChunkLineSelector", lines, false);
-  for (auto&& line : lines)
+  rcg::getEnum(rcgnodemap, "LineSelector", lines, true);
+  uint32_t input_mask = 0;
+  uint32_t output_mask = 0;
+  for (size_t i = 0; i < lines.size(); i++)
   {
-    rc_common_msgs::KeyValue line_source;
-    line_source.key = line;
-    rcg::setEnum(rcgnodemap, "ChunkLineSelector", line.c_str(), false);       // set respective selector
-    line_source.value = rcg::getEnum(rcgnodemap, "ChunkLineSource", false);   // get the actual source value
-    cam_param.line_source.push_back(line_source);
+    rcg::setEnum(rcgnodemap, "LineSelector", lines[i].c_str(), true);
+    std::string io = rcg::getEnum(rcgnodemap, "LineMode");
+
+    if (io == "Input")
+    {
+      input_mask |= 1 << i;
+    }
+
+    if (io == "Output")
+    {
+      output_mask |= 1 << i;
+
+      // for output also get LineSource for this image
+      rcg::setEnum(rcgnodemap, "ChunkLineSelector", lines[i].c_str(), true);
+      rc_common_msgs::KeyValue line_source;
+      line_source.key = lines[i];
+      line_source.value = rcg::getEnum(rcgnodemap, "ChunkLineSource", true);
+      cam_param.line_source.push_back(line_source);
+    }
   }
+  cam_param.extra_data.push_back(getKeyValue("input_mask", input_mask));
+  cam_param.extra_data.push_back(getKeyValue("output_mask", output_mask));
 
   // get LineStatusAll
   cam_param.line_status_all = rcg::getInteger(rcgnodemap, "ChunkLineStatusAll", 0, 0, false);
@@ -1277,68 +1365,70 @@ rc_common_msgs::CameraParam extractChunkData(const std::shared_ptr<GenApi::CNode
   cam_param.exposure_time = rcg::getFloat(rcgnodemap, "ChunkExposureTime", 0, 0, false) / 1000000l;
 
   // add extra data
-  cam_param.extra_data.clear();
   try
   {
-    rc_common_msgs::KeyValue kv;
-    kv.key = "noise";
-    kv.value = std::to_string(rcg::getFloat(rcgnodemap, "ChunkRcNoise", 0, 0, true));
-    cam_param.extra_data.push_back(kv);
+    float noise = rcg::getFloat(rcgnodemap, "ChunkRcNoise", 0, 0, true);
+    cam_param.extra_data.push_back(getKeyValue("noise", noise));
   }
-  catch (std::invalid_argument& e)
+  catch (std::invalid_argument&)
   {
     // calculate camera's noise level
     static float NOISE_BASE = 2.0;
-    rc_common_msgs::KeyValue kv;
-    kv.key = "noise";
-    kv.value = std::to_string(static_cast<float>(NOISE_BASE * std::exp(cam_param.gain * std::log(10)/20)));
-    cam_param.extra_data.push_back(kv);
+    float noise = static_cast<float>(NOISE_BASE * std::exp(cam_param.gain * std::log(10)/20));
+    cam_param.extra_data.push_back(getKeyValue("noise", noise));
   }
 
   try
   {
-    rc_common_msgs::KeyValue kv;
-    kv.key = "test";
-    kv.value = std::to_string(rcg::getBoolean(rcgnodemap, "ChunkRcTest", true));
-    cam_param.extra_data.push_back(kv);
+    bool test = rcg::getBoolean(rcgnodemap, "ChunkRcTest", true);
+    cam_param.extra_data.push_back(getKeyValue("test", test));
   }
-  catch (std::invalid_argument& e)
+  catch (std::invalid_argument&)
   {
+    // feature not available on device
   }
 
   try
   {
-    std::string out1_reduction;
+    float out1_reduction;
     try
     {
-      out1_reduction = std::to_string(rcg::getFloat(rcgnodemap, "ChunkRcOut1Reduction", 0, 0, true));
+      out1_reduction = rcg::getFloat(rcgnodemap, "ChunkRcOut1Reduction", 0, 0, true);
     }
-    catch (const std::exception& e)
+    catch (const std::exception&)
     {
       // try to fallback to older name in versions < 20.10.1
-      out1_reduction = std::to_string(rcg::getFloat(rcgnodemap, "ChunkRcAdaptiveOut1Reduction", 0, 0, true));
+      out1_reduction = rcg::getFloat(rcgnodemap, "ChunkRcAdaptiveOut1Reduction", 0, 0, true);
     }
-
-    rc_common_msgs::KeyValue kv;
-    kv.key = "out1_reduction";
-    kv.value = out1_reduction;
-    cam_param.extra_data.push_back(kv);
+    cam_param.extra_data.push_back(getKeyValue("out1_reduction", out1_reduction));
   }
-  catch (std::invalid_argument& e)
+  catch (std::invalid_argument&)
   {
+    // feature not available on device
   }
 
   try
   {
-    rc_common_msgs::KeyValue kv;
-    kv.key = "brightness";
-    kv.value = std::to_string(rcg::getFloat(rcgnodemap, "ChunkRcBrightness", 0, 0, true));
-    cam_param.extra_data.push_back(kv);
+    float brightness = rcg::getFloat(rcgnodemap, "ChunkRcBrightness", 0, 0, true);
+    cam_param.extra_data.push_back(getKeyValue("brightness", brightness));
   }
-  catch (std::invalid_argument& e)
+  catch (std::invalid_argument&)
   {
+    // feature not available on device
   }
 
+  try
+  {
+    bool adapting = rcg::getBoolean(rcgnodemap, "ChunkRcAutoExposureAdapting", true);
+    cam_param.extra_data.push_back(getKeyValue("auto_exposure_adapting", adapting));
+  }
+  catch (const std::exception&)
+  {
+    // feature not available on device
+  }
+
+  cam_param.extra_data.push_back(getKeyValueString("model_name", rcg::getString(rcgnodemap, "DeviceModelName")));
+
   return cam_param;
 }
 
@@ -1430,10 +1520,7 @@ void DeviceNodelet::grab(std::string device, rcg::Device::ACCESS access)
       // prepare chunk adapter for getting chunk data
 
       std::shared_ptr<GenApi::CChunkAdapter> chunkadapter;
-      if (dev_supports_chunk_data)
-      {
-        chunkadapter = rcg::getChunkAdapter(rcgnodemap, rcgdev->getTLType());
-      }
+      chunkadapter = rcg::getChunkAdapter(rcgnodemap, rcgdev->getTLType());
 
       // initialize all publishers
 
@@ -1468,13 +1555,8 @@ void DeviceNodelet::grab(std::string device, rcg::Device::ACCESS access)
 
       // add camera/image params publishers if the camera supports chunkdata
 
-      std::shared_ptr<CameraParamPublisher> lcamparams;
-      std::shared_ptr<CameraParamPublisher> rcamparams;
-      if (dev_supports_chunk_data)
-      {
-        lcamparams = std::make_shared<CameraParamPublisher>(nh, tfPrefix, true);
-        rcamparams = std::make_shared<CameraParamPublisher>(nh, tfPrefix, false);
-      }
+      std::shared_ptr<CameraParamPublisher> lcamparams = std::make_shared<CameraParamPublisher>(nh, tfPrefix, true);
+      std::shared_ptr<CameraParamPublisher> rcamparams = std::make_shared<CameraParamPublisher>(nh, tfPrefix, false);
 
       // start streaming of first stream
 
@@ -1530,7 +1612,17 @@ void DeviceNodelet::grab(std::string device, rcg::Device::ACCESS access)
             bool out1_alternate = (out1_mode_on_sensor == "ExposureAlternateActive");
             points2.setOut1Alternate(out1_alternate);
 
-            rc_common_msgs::CameraParam cam_param = extractChunkData(rcgnodemap);
+            rc_common_msgs::CameraParam cam_param;
+
+            {
+              // protect against concurent changes of selector (i.e. LineSelector),
+              // before reading the corresponding value
+
+              boost::recursive_mutex::scoped_lock lock(mtx);
+
+              cam_param = extractChunkData(rcgnodemap);
+            }
+
             cam_param.is_color_camera = dev_supports_color;
             out1 = (cam_param.line_status_all & 0x1);
 
@@ -1651,7 +1743,10 @@ void DeviceNodelet::grab(std::string device, rcg::Device::ACCESS access)
             if (!ccolor)
             {
               rcg::setEnum(rcgnodemap, "ComponentSelector", "Intensity", true);
-              rcg::setEnum(rcgnodemap, "PixelFormat", "YCbCr411_8", true);
+              rcg::setEnum(rcgnodemap, "PixelFormat", color_format.c_str(), false);
+              rcg::setEnum(rcgnodemap, "ComponentSelector", "IntensityCombined", true);
+              rcg::setEnum(rcgnodemap, "PixelFormat", color_format.c_str(), false);
+
               ccolor = true;
             }
           }
